/*
 * rctl -- Digium R-Series control application
 *
 * Copyright (C) 2011, Digium, Inc.
 *
 * Jason Parker <qwell@digium.com>
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2.  See the COPYING file
 * at the top of the source tree.
 */

#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <inttypes.h>
#include <poll.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>

#define CMD_REQ(cmd) \
	struct R_cmd_req_##cmd

#define CMD_RESP(cmd) \
	struct R_cmd_resp_##cmd

#define DEF_CMD(cmd, req, resp) \
	CMD_REQ(cmd) { \
		req \
	} __attribute__ ((__packed__)); \
	CMD_RESP(cmd) { \
		resp \
	} __attribute__ ((__packed__));

DEF_CMD(CMD_GET_FIRMWARE_VERSION,
	,
	uint16_t firmware_version;
);

DEF_CMD(CMD_GET_SERIAL_NUMBER,
	,
	char serial_number[20];
);

DEF_CMD(CMD_SET_SERIAL_NUMBER,
	char serial_number[20];
	uint32_t digium_magic;
	,
	uint8_t ack;
);

DEF_CMD(CMD_GET_PRODUCT_NUMBER,
	,
	char product_number[32];
);

DEF_CMD(CMD_SET_PRODUCT_NUMBER,
	char product_number[32];
	,
	uint8_t ack;
);

DEF_CMD(CMD_GET_PORT_COUNT,
	,
	uint8_t count;
);

DEF_CMD(CMD_GET_CURRENT_CTLPORT,
	,
	uint8_t port;
);

DEF_CMD(CMD_GET_MASTER_CTLPORT,
	,
	uint8_t port;
);

DEF_CMD(CMD_GET_ID_SWITCH,
	,
	uint8_t id;
);

DEF_CMD(CMD_FIRMWARE_STAGE1,
	,
	uint8_t ack;
);

DEF_CMD(CMD_FIRMWARE_STAGE2,
	,
	uint8_t ack;
);

DEF_CMD(CMD_GET_WATCHDOG,
	uint8_t port;
	,
	uint8_t port;
	uint8_t timeout;
);

DEF_CMD(CMD_SET_WATCHDOG,
	uint8_t port;
	uint8_t timeout;
	,
	uint8_t port;
	uint8_t timeout;
);

DEF_CMD(CMD_PING_WATCHDOG,
	,
	uint8_t ack;
);

DEF_CMD(CMD_CONTROL_REQUEST,
	,
	uint8_t result;
);

DEF_CMD(CMD_CONTROL_RELEASE,
	,
	uint8_t result;
);

DEF_CMD(CMD_CONTROL_CHANGE,
	,
	uint8_t port;
);

DEF_CMD(CMD_GET_CONTROL_MODE,
	,
	uint8_t mode;
);

DEF_CMD(CMD_SET_CONTROL_MODE,
	uint8_t mode;
	,
	uint8_t ack;
);

struct portinfo {
	uint8_t port;
	uint8_t type;
	uint8_t passthrough;
	uint8_t status;	
} __attribute__ ((__packed__));

DEF_CMD(CMD_GET_PORT_STATUS,
	,
	struct portinfo portlist[8];
);

DEF_CMD(CMD_SET_PORT_STATUS,
	struct portinfo portlist[8];
	,
	uint8_t ack;
);

struct R_packet {
	uint8_t len;
	uint8_t cmd;
	char data[253];
} __attribute__ ((__packed__));

enum {
	CMD_GET_FIRMWARE_VERSION = 0,
	CMD_GET_SERIAL_NUMBER,
	CMD_SET_SERIAL_NUMBER,
	CMD_GET_PRODUCT_NUMBER,
	CMD_SET_PRODUCT_NUMBER,
	CMD_GET_PORT_COUNT,
	CMD_GET_CURRENT_CTLPORT,
	CMD_GET_MASTER_CTLPORT,
	CMD_GET_ID_SWITCH,
	CMD_FIRMWARE_STAGE1,
	CMD_FIRMWARE_STAGE2,
	CMD_GET_WATCHDOG,
	CMD_SET_WATCHDOG,
	CMD_PING_WATCHDOG,
	CMD_CONTROL_REQUEST,
	CMD_CONTROL_RELEASE,
	CMD_CONTROL_CHANGE,
	CMD_GET_CONTROL_MODE,
	CMD_SET_CONTROL_MODE,
	CMD_GET_PORT_STATUS,
	CMD_SET_PORT_STATUS,
};

struct R_command {
	const uint8_t cmd;  /* Command code as defined by R-Series API. */
	const char *option; /* Name of CLI option.  See long_options[] in main() */
};

const struct R_command R_commands[] = {
	[CMD_GET_FIRMWARE_VERSION] = {(0x02), "get-firmware-version"},
	[CMD_GET_SERIAL_NUMBER]    = {(0x03), "get-serial-number"},
	[CMD_SET_SERIAL_NUMBER]    = {(0x04), "set-serial-number"},
	[CMD_GET_PRODUCT_NUMBER]   = {(0x05), "get-product-number"},
	[CMD_SET_PRODUCT_NUMBER]   = {(0x06), "set-product-number"},
	[CMD_GET_PORT_COUNT]       = {(0x08), "get-port-count"},
	[CMD_GET_CURRENT_CTLPORT]  = {(0x09), "get-current-ctlport"},
	[CMD_GET_MASTER_CTLPORT]   = {(0x0A), "get-master-ctlport"},
	[CMD_GET_ID_SWITCH]        = {(0x0B), "get-id-switch"},
	[CMD_FIRMWARE_STAGE1]      = {(0x0C), "update-firmware"}, /* Set twice, to force both to be called. */
	[CMD_FIRMWARE_STAGE2]      = {(0x0D), "update-firmware"},
	[CMD_GET_WATCHDOG]         = {(0x17), "get-watchdog"},
	[CMD_SET_WATCHDOG]         = {(0x15), "set-watchdog"},
	[CMD_PING_WATCHDOG]        = {(0x19), "ping-watchdog"},
	[CMD_CONTROL_REQUEST]      = {(0x1A), "control-request"},
	[CMD_CONTROL_RELEASE]      = {(0x1B), "control-release"},
	[CMD_CONTROL_CHANGE]       = {(0x1C), NULL},
	[CMD_GET_CONTROL_MODE]     = {(0x1D), "get-control-mode"},
	[CMD_SET_CONTROL_MODE]     = {(0x1E), "set-control-mode"},
	[CMD_GET_PORT_STATUS]      = {(0x28), "get-port-status"},
	[CMD_SET_PORT_STATUS]      = {(0x27), "set-port-status"},
};

#define ACK 0x06
#define NAK 0x15
enum R_errors {
	ERR_FAILURE           = (NAK - 0x00),
};

const char *error2str(enum R_errors error) {
	switch (error) {
		case ERR_FAILURE:
			return "Failed.";
		default:
			return "An unknown error occured.";
	}
}

enum R_port_types {
	PORT_TYPE_T1E1     = (0x00),
	PORT_TYPE_BRI      = (0x01),
	PORT_TYPE_ANALOG   = (0x02),
	PORT_TYPE_NOCHANGE = (0xFF),
};

const char *porttype2str(enum R_port_types type) {
	switch (type) {
		case PORT_TYPE_T1E1:
			return "T1/E1";
		case PORT_TYPE_BRI:
			return "BRI";
		case PORT_TYPE_ANALOG:
			return "analog";
		case PORT_TYPE_NOCHANGE:
			return "no change";
		default:
			return "unknown";
	}
}

enum R_port_statuses {
	PORT_STATUS_PRIMARY   = (0x00),
	PORT_STATUS_SECONDARY = (0x01),
	PORT_STATUS_NOCHANGE  = (0xFF),
};

const char *portstatus2str(enum R_port_statuses status) {
	switch (status) {
		case PORT_STATUS_PRIMARY:
			return "primary";
		case PORT_STATUS_SECONDARY:
			return "secondary";
		case PORT_STATUS_NOCHANGE:
			return "no change";
		default:
		{
			char buf[16];
			sprintf(buf, "unknown (%d)", status);
			return strdup(buf);
		}
	}
}

enum R_port_passthroughs {
	PORT_PASSTHROUGH_DISABLED = (0x00),
	PORT_PASSTHROUGH_ENABLED  = (0x01),
	PORT_PASSTHROUGH_NOCHANGE = (0xFF),
};

int ttyfd;
static int debugmode = 0;
static uint16_t ports = 0; /* Bitmask of ports to use for get/set port status. */
static uint32_t cmd_mask = 0;
static char *firmware;

struct R_packet *create_packet(void **reqresp) {
	struct R_packet *pkt;
	pkt = calloc(1, sizeof(*pkt));

	if (reqresp) {
		*reqresp = (void *)pkt->data;
	}
	return pkt;
}

void *destroy_packet(struct R_packet *pkt) {
	free(pkt);
}

void *R_log(const char *format, ...) {
	va_list va;
	va_start(va, format);
	vfprintf(stdout, format, va);
	va_end(va);
}

void *R_err(const char *format, ...) {
	va_list va;
	va_start(va, format);
	vfprintf(stderr, format, va);
	va_end(va);
}

void *R_debug(const char *format, ...) {
	va_list va;

	if (!debugmode) {
		return;
	}

	va_start(va, format);
	vfprintf(stderr, format, va);
	va_end(va);
}

int parse_ack(uint8_t ackcode) {
	int ret;
	switch (ackcode) {
		case ACK:
			ret = 1;
			break;
		case ERR_FAILURE:
		default:
			R_err("Error executing command: %s\n", error2str(ackcode));
			ret = 0;
			break;
	}

	return ret;
}

void update_firmware() {
	if (isp_synch()) {
		R_err("Could not synch to device\n");
	} else {
		FILE *f;
		R_log("Synchronized\n");
		if ((f = fopen(firmware, "r"))) {
			int start = 0xD100;
			int len;
			int i;
			uint8_t *buffer;
			int prog_version = 0x0101;
			char *prog_id = NULL;

			prog_id = basename(firmware);
			fseek(f, 0L, SEEK_END);
			len = ftell(f);
			fseek(f, 0L, SEEK_SET);
			buffer = malloc(len);
			if (buffer) {
				len = fread(buffer, 1, len, f);
			}
			fclose(f);

			/* This may not be working properly.  Disable for now.
			isp_setEcho(0);
			*/
			writeEncryptedImage(buffer, len, prog_version, prog_id);
			if (!(i = isp_doG(start))) {
				R_log("Executing at address 0x%x\n", start);
			} else {
				R_err("Unable to execute at address 0x%x (%d -%s)\n", start, i, isp_response_string(i));
			}
		}
	}
}

int send_request(struct R_packet *pkt, size_t reqsize) {
	int res;
	int totallen;
	int len;

	pkt->len = sizeof(pkt->len) + sizeof(pkt->cmd) + reqsize;
	R_debug("req - pkt->len: %d, pkt->cmd: %d, pkt->data: %s\n", pkt->len, pkt->cmd, pkt->data);

	len = totallen = pkt->len;

	while ((res = write(ttyfd, pkt + (totallen - len), len)) <= (len - 1)) {
		if (res < 0) {
			if (errno == EAGAIN) {
				continue;
			}

			return 1;
		}
		len -= res;
	}

	return 0;
}

int handle_response() {
	struct R_packet *pkt;
	int i;
	int res;
	int cmd = -1;

	pkt = create_packet(NULL);
	while ((res = read(ttyfd, &pkt->len, 1)) <= 0) {
		if (res < 0) {
			if (errno == EAGAIN) {
				continue;
			}

			return 1;
		}
	}

	if (pkt->len > 1) {
		while ((res = read(ttyfd, &pkt->cmd, 1)) <= 0) {
			if (res < 0) {
				if (errno == EAGAIN) {
					continue;
				}

				return 1;
			}
		}
	} else {
		return 1;
	}

	if (pkt->len > 2) {
		int totallen;
		int len;
		len = totallen = (pkt->len - 2);

		while ((res = read(ttyfd, &pkt->data + (totallen - len), len)) <= (len - 1)) {
			if (res < 0) {
				if (errno == EAGAIN) {
					continue;
				}

				return 1;
			}
			len -= res;
		}
	}

	for (i = 0; i < (sizeof(R_commands) / sizeof(struct R_command)); i++) {
		if (pkt->cmd == R_commands[i].cmd) {
			cmd = i;
			break;
		}
	}

	if (cmd_mask & (1 << cmd)) {
		cmd_mask &= ~(1 << cmd);
		switch (cmd) {
			case CMD_GET_FIRMWARE_VERSION:
				response_get_firmware_version(pkt);
				break;
			case CMD_GET_SERIAL_NUMBER:
				response_get_serial_number(pkt);
				break;
			case CMD_SET_SERIAL_NUMBER:
				response_set_serial_number(pkt);
				break;
			case CMD_GET_PRODUCT_NUMBER:
				response_get_product_number(pkt);
				break;
			case CMD_SET_PRODUCT_NUMBER:
				response_set_product_number(pkt);
				break;
			case CMD_GET_PORT_COUNT:
				response_get_port_count(pkt);
				break;
			case CMD_GET_CURRENT_CTLPORT:
				response_get_current_ctlport(pkt);
				break;
			case CMD_GET_MASTER_CTLPORT:
				response_get_master_ctlport(pkt);
				break;
			case CMD_GET_ID_SWITCH:
				response_get_id_switch(pkt);
				break;
			case CMD_FIRMWARE_STAGE1:
				response_firmware_stage1(pkt);
				break;
			case CMD_FIRMWARE_STAGE2:
				response_firmware_stage2(pkt);
				break;
			case CMD_GET_WATCHDOG:
				response_get_watchdog(pkt);
				break;
			case CMD_SET_WATCHDOG:
				response_set_watchdog(pkt);
				break;
			case CMD_PING_WATCHDOG:
				response_ping_watchdog(pkt);
				break;
			case CMD_CONTROL_REQUEST:
				response_control_request(pkt);
				break;
			case CMD_CONTROL_RELEASE:
				response_control_release(pkt);
				break;
			case CMD_CONTROL_CHANGE:
				response_control_change(pkt);
				break;
			case CMD_GET_CONTROL_MODE:
				response_get_control_mode(pkt);
				break;
			case CMD_SET_CONTROL_MODE:
				response_set_control_mode(pkt);
				break;
			case CMD_GET_PORT_STATUS:
				response_get_port_status(pkt);
				break;
			case CMD_SET_PORT_STATUS:
				response_set_port_status(pkt);
				break;
			default:
				R_err("Got a response to command '%d' which we don't understand.\n", pkt->cmd);
				break;
		}
	} else {
		R_err("Got unsolicited response '%d'.\n", pkt->cmd);
	}
	R_debug("resp - pkt->len: %d, pkt->cmd: %d, pkt->data: %s\n", pkt->len, pkt->cmd, pkt->data);
	return 0;
}

int request_get_firmware_version() {
	struct R_packet *pkt;
	CMD_REQ(CMD_GET_FIRMWARE_VERSION) *req;
	int ret;

	pkt = create_packet((void *)&req);
	pkt->cmd = R_commands[CMD_GET_FIRMWARE_VERSION].cmd;
	ret = send_request(pkt, sizeof(*req));
	destroy_packet(pkt);
	return ret;
}

int response_get_firmware_version(struct R_packet *pkt) {
	CMD_RESP(CMD_GET_FIRMWARE_VERSION) *resp = (void *)pkt->data;

	R_log("Firmware version: %d\n", ntohs(resp->firmware_version));
	return 0;
}

int request_get_serial_number() {
	struct R_packet *pkt;
	CMD_REQ(CMD_GET_SERIAL_NUMBER) *req;
	int ret;

	pkt = create_packet((void *)&req);
	pkt->cmd = R_commands[CMD_GET_SERIAL_NUMBER].cmd;
	ret = send_request(pkt, sizeof(*req));
	destroy_packet(pkt);
	return ret;
}

int response_get_serial_number(struct R_packet *pkt) {
	CMD_RESP(CMD_GET_SERIAL_NUMBER) *resp = (void *)pkt->data;

	R_log("Serial number: %s\n", resp->serial_number);
	return 0;
}

int request_set_serial_number(const char *serial_number) {
	struct R_packet *pkt;
	CMD_REQ(CMD_SET_SERIAL_NUMBER) *req;
	int ret;

	pkt = create_packet((void *)&req);
	pkt->cmd = R_commands[CMD_SET_SERIAL_NUMBER].cmd;
	strcpy(req->serial_number, serial_number);
	req->digium_magic = htonl(0xFFFF);
	ret = send_request(pkt, sizeof(*req));
	destroy_packet(pkt);
	return ret;
}

int response_set_serial_number(struct R_packet *pkt) {
	CMD_RESP(CMD_SET_SERIAL_NUMBER) *resp = (void *)pkt->data;

	if (parse_ack(resp->ack)) {
		R_debug("Serial number set successfully.\n");
	}
	return 0;
}

int request_get_product_number() {
	struct R_packet *pkt;
	CMD_REQ(CMD_GET_PRODUCT_NUMBER) *req;
	int ret;

	pkt = create_packet((void *)&req);
	pkt->cmd = R_commands[CMD_GET_PRODUCT_NUMBER].cmd;
	ret = send_request(pkt, sizeof(*req));
	destroy_packet(pkt);
	return ret;
}

int response_get_product_number(struct R_packet *pkt) {
	CMD_RESP(CMD_GET_PRODUCT_NUMBER) *resp = (void *)pkt->data;

	R_log("Product number: %s\n", resp->product_number);
	return 0;
}

int request_set_product_number(const char *product_number) {
	struct R_packet *pkt;
	CMD_REQ(CMD_SET_PRODUCT_NUMBER) *req;
	int ret;

	pkt = create_packet((void *)&req);
	pkt->cmd = R_commands[CMD_SET_PRODUCT_NUMBER].cmd;
	strcpy(req->product_number, product_number);
	ret = send_request(pkt, sizeof(*req));
	destroy_packet(pkt);
	return ret;
}

int response_set_product_number(struct R_packet *pkt) {
	CMD_RESP(CMD_SET_PRODUCT_NUMBER) *resp = (void *)pkt->data;

	if (parse_ack(resp->ack)) {
		R_debug("Product number set successfully.\n");
	}
	return 0;
}

int request_get_port_count() {
	struct R_packet *pkt;
	CMD_REQ(CMD_GET_PORT_COUNT) *req;
	int ret;

	pkt = create_packet((void *)&req);
	pkt->cmd = R_commands[CMD_GET_PORT_COUNT].cmd;
	ret = send_request(pkt, sizeof(*req));
	destroy_packet(pkt);
	return ret;
}

int response_get_port_count(struct R_packet *pkt) {
	CMD_RESP(CMD_GET_PORT_COUNT) *resp = (void *)pkt->data;

	R_log("Ports: %d\n", resp->count);
	return 0;
}

int request_get_current_ctlport() {
	struct R_packet *pkt;
	CMD_REQ(CMD_GET_CURRENT_CTLPORT) *req;
	int ret;

	pkt = create_packet((void *)&req);
	pkt->cmd = R_commands[CMD_GET_CURRENT_CTLPORT].cmd;
	ret = send_request(pkt, sizeof(*req));
	destroy_packet(pkt);
	return ret;
}

int response_get_current_ctlport(struct R_packet *pkt) {
	CMD_RESP(CMD_GET_CURRENT_CTLPORT) *resp = (void *)pkt->data;

	R_log("Current control port: %d\n", resp->port);
	return 0;
}

int request_get_master_ctlport() {
	struct R_packet *pkt;
	CMD_REQ(CMD_GET_MASTER_CTLPORT) *req;
	int ret;

	pkt = create_packet((void *)&req);
	pkt->cmd = R_commands[CMD_GET_MASTER_CTLPORT].cmd;
	ret = send_request(pkt, sizeof(*req));
	destroy_packet(pkt);
	return ret;
}

int response_get_master_ctlport(struct R_packet *pkt) {
	CMD_RESP(CMD_GET_MASTER_CTLPORT) *resp = (void *)pkt->data;

	R_log("Master control port: %d\n", resp->port);
	return 0;
}

int request_get_id_switch() {
	struct R_packet *pkt;
	CMD_REQ(CMD_GET_ID_SWITCH) *req;
	int ret;

	pkt = create_packet((void *)&req);
	pkt->cmd = R_commands[CMD_GET_ID_SWITCH].cmd;
	ret = send_request(pkt, sizeof(*req));
	destroy_packet(pkt);
	return ret;
}

int response_get_id_switch(struct R_packet *pkt) {
	CMD_RESP(CMD_GET_ID_SWITCH) *resp = (void *)pkt->data;

	R_log("ID Switch: %d\n", resp->id);
	return 0;
}

int request_firmware_stage1() {
	struct R_packet *pkt;
	CMD_REQ(CMD_FIRMWARE_STAGE1) *req;
	int ret;

	pkt = create_packet((void *)&req);
	pkt->cmd = R_commands[CMD_FIRMWARE_STAGE1].cmd;
	ret = send_request(pkt, sizeof(*req));
	destroy_packet(pkt);
	return ret;
}

int response_firmware_stage1(struct R_packet *pkt) {
	CMD_RESP(CMD_FIRMWARE_STAGE1) *resp = (void *)pkt->data;

	if (parse_ack(resp->ack)) {
		R_log("Firmware stage1 completed successfully.\n");
	}
	return 0;
}

int request_firmware_stage2() {
	struct R_packet *pkt;
	CMD_REQ(CMD_FIRMWARE_STAGE2) *req;
	int ret;

	pkt = create_packet((void *)&req);
	pkt->cmd = R_commands[CMD_FIRMWARE_STAGE2].cmd;
	ret = send_request(pkt, sizeof(*req));
	destroy_packet(pkt);
	return ret;
}

int response_firmware_stage2(struct R_packet *pkt) {
	CMD_RESP(CMD_FIRMWARE_STAGE2) *resp = (void *)pkt->data;

	if (parse_ack(resp->ack)) {
		R_log("Firmware stage2 completed successfully.\n");
		update_firmware();
	}
	return 0;
}

int request_get_watchdog() {
	struct R_packet *pkt;
	CMD_REQ(CMD_GET_WATCHDOG) *req;
	int ret;

	pkt = create_packet((void *)&req);
	pkt->cmd = R_commands[CMD_GET_WATCHDOG].cmd;
	req->port = 0;
	ret = send_request(pkt, sizeof(*req));
	destroy_packet(pkt);
	return ret;
}

int response_get_watchdog(struct R_packet *pkt) {
	CMD_RESP(CMD_GET_WATCHDOG) *resp = (void *)pkt->data;

	R_log("Watchdog Timeout (get): %d\n", resp->timeout);
	return 0;
}

int request_set_watchdog(uint8_t timeout) {
	struct R_packet *pkt;
	CMD_REQ(CMD_SET_WATCHDOG) *req;
	int ret;

	pkt = create_packet((void *)&req);
	pkt->cmd = R_commands[CMD_SET_WATCHDOG].cmd;
	req->port = 0;
	req->timeout = timeout;
	ret = send_request(pkt, sizeof(*req));
	destroy_packet(pkt);
	return ret;
}

int response_set_watchdog(struct R_packet *pkt) {
	CMD_RESP(CMD_SET_WATCHDOG) *resp = (void *)pkt->data;

	R_log("Watchdog Timeout (set): %d\n", resp->timeout);
	return 0;
}

int request_ping_watchdog() {
	struct R_packet *pkt;
	CMD_REQ(CMD_PING_WATCHDOG) *req;
	int ret;

	pkt = create_packet((void *)&req);
	pkt->cmd = R_commands[CMD_PING_WATCHDOG].cmd;
	ret = send_request(pkt, sizeof(*req));
	destroy_packet(pkt);
	return ret;
}

int response_ping_watchdog(struct R_packet *pkt) {
	CMD_RESP(CMD_PING_WATCHDOG) *resp = (void *)pkt->data;

	if (parse_ack(resp->ack)) {
		R_log("PONG\n");
	}
	return 0;
}

int request_control_request() {
	struct R_packet *pkt;
	CMD_REQ(CMD_CONTROL_REQUEST) *req;
	int ret;

	pkt = create_packet((void *)&req);
	pkt->cmd = R_commands[CMD_CONTROL_REQUEST].cmd;
	ret = send_request(pkt, sizeof(*req));
	destroy_packet(pkt);
	return ret;
}

int response_control_request(struct R_packet *pkt) {
	CMD_RESP(CMD_CONTROL_REQUEST) *resp = (void *)pkt->data;

	R_log("Control Request Result: %d\n", resp->result);
	return 0;
}

int control_request_release() {
	struct R_packet *pkt;
	CMD_REQ(CMD_CONTROL_RELEASE) *req;
	int ret;

	pkt = create_packet((void *)&req);
	pkt->cmd = R_commands[CMD_CONTROL_RELEASE].cmd;
	ret = send_request(pkt, sizeof(*req));
	destroy_packet(pkt);
	return ret;
}

int response_control_release(struct R_packet *pkt) {
	CMD_RESP(CMD_CONTROL_RELEASE) *resp = (void *)pkt->data;

	R_log("Control Release Result: %d\n", resp->result);
	return 0;
}

int response_control_change(struct R_packet *pkt) {
	CMD_RESP(CMD_CONTROL_CHANGE) *resp = (void *)pkt->data;

	R_log("Control Changed to Port: %d\n", resp->port);
	return 0;
}

int request_get_control_mode() {
	struct R_packet *pkt;
	CMD_REQ(CMD_GET_CONTROL_MODE) *req;
	int ret;

	pkt = create_packet((void *)&req);
	pkt->cmd = R_commands[CMD_GET_CONTROL_MODE].cmd;
	ret = send_request(pkt, sizeof(*req));
	destroy_packet(pkt);
	return ret;
}

int response_get_control_mode(struct R_packet *pkt) {
	CMD_RESP(CMD_GET_CONTROL_MODE) *resp = (void *)pkt->data;

	R_log("Control mode: %d\n", resp->mode);

	return 0;
}

int request_set_control_mode(uint8_t mode) {
	struct R_packet *pkt;
	CMD_REQ(CMD_SET_CONTROL_MODE) *req;
	int ret;
	int i;

	pkt = create_packet((void *)&req);
	pkt->cmd = R_commands[CMD_SET_CONTROL_MODE].cmd;

	req->mode = mode;

	ret = send_request(pkt, sizeof(*req));
	destroy_packet(pkt);
	return ret;
}

int response_set_control_mode(struct R_packet *pkt) {
	CMD_RESP(CMD_SET_CONTROL_MODE) *resp = (void *)pkt->data;

	if (parse_ack(resp->ack)) {
		R_debug("Control mode set successfully.\n");
	}
	return 0;
}

int request_get_port_status() {
	struct R_packet *pkt;
	CMD_REQ(CMD_GET_PORT_STATUS) *req;
	int ret;

	pkt = create_packet((void *)&req);
	pkt->cmd = R_commands[CMD_GET_PORT_STATUS].cmd;
	ret = send_request(pkt, sizeof(*req));
	destroy_packet(pkt);
	return ret;
}

int response_get_port_status(struct R_packet *pkt) {
	CMD_RESP(CMD_GET_PORT_STATUS) *resp = (void *)pkt->data;
	int i;

	if (ports > 0) {
		for (i = 0; i < ((pkt->len - 2) / sizeof(struct portinfo)); i++) {
			resp->portlist[i].port = i + 1;
			if (ports & (1 << i)) {
				R_log("Port %d: %d, %s, %s\n", resp->portlist[i].port, resp->portlist[i].passthrough, portstatus2str(resp->portlist[i].status), porttype2str(resp->portlist[i].type));
			}
		}
	}
	return 0;
}

int request_set_port_status(uint8_t status) {
	struct R_packet *pkt;
	CMD_REQ(CMD_SET_PORT_STATUS) *req;
	int ret;
	int i;

	pkt = create_packet((void *)&req);
	pkt->cmd = R_commands[CMD_SET_PORT_STATUS].cmd;

	if (ports > 0) {
		for (i = 0; i <= 7; i++) {
			req->portlist[i].port = i + 1;
			req->portlist[i].passthrough = PORT_PASSTHROUGH_NOCHANGE;
			req->portlist[i].type = PORT_TYPE_NOCHANGE;
			if (ports & (1 << i)) {
				req->portlist[i].status = status;
			} else {
				req->portlist[i].status = PORT_STATUS_NOCHANGE;
			}
		}
	}
	ret = send_request(pkt, sizeof(*req));
	destroy_packet(pkt);
	return ret;
}

int response_set_port_status(struct R_packet *pkt) {
	CMD_RESP(CMD_SET_PORT_STATUS) *resp = (void *)pkt->data;

	if (parse_ack(resp->ack)) {
		R_debug("Port(s) set successfully.\n");
	}
	return 0;
}

#define R_usage() \
	R_log("Usage: %s [-d] [-p port[,port[,port[,...]]]] [-s switchid] [-t device] [-w]\n" \
	      "  Long options:\n" \
	      "    --get-firmware-version\n" \
	      "    --get-serial-number\n" \
	      "    --set-serial-number     SN\n" \
	      "    --get-product-number\n" \
	      "    --set-product-number    PN\n" \
	      "    --get-port-count\n" \
	      "    --get-current-ctlport\n" \
	      "    --get-master-ctlport\n" \
	      "    --get-id-switch\n" \
	      "    --update-firmware       filename\n" \
	      "    --get-watchdog\n" \
	      "    --set-watchdog\n" \
	      "    --ping-watchdog\n" \
	      "    --control-request\n" \
	      "    --control-release\n" \
	      "    --get-control-mode\n" \
	      "    --set-control-mode      0|1\n" \
	      "    --get-port-status\n" \
	      "    --set-port-status       0|1\n" , argv[0]);

int main(int argc, char *argv[]) {
	struct pollfd fds;
	int timeout = 1000;
	int option_index = 0;
	int opt;
	int i;
	struct termios termios_p;

	char *usbtty = NULL;

	char *optargs[(sizeof(R_commands) / sizeof(struct R_command))];

	const struct option long_options[] = {
		/* Don't forget to add to R_commands[].option */
		{"get-firmware-version", 0, NULL, 0},
		{"get-serial-number", 0, NULL, 0},
		{"set-serial-number", 1, NULL, 0},
		{"get-product-number", 0, NULL, 0},
		{"set-product-number", 1, NULL, 0},
		{"get-port-count", 0, NULL, 0},
		{"get-current-ctlport", 0, NULL, 0},
		{"get-master-ctlport", 0, NULL, 0},
		{"get-id-switch", 0, NULL, 0},
		{"update-firmware", 1, NULL, 0},
		{"get-watchdog", 0, NULL, 0},
		{"set-watchdog", 1, NULL, 0},
		{"ping-watchdog", 0, NULL, 0},
		{"control-request", 0, NULL, 0},
		{"control-release", 0, NULL, 0},
		{"get-control-mode", 0, NULL, 0},
		{"set-control-mode", 1, NULL, 0},
		{"get-port-status", 0, NULL, 0},
		{"set-port-status", 1, NULL, 0},

		{"debug", 0, NULL, 'd'},
		{"help", 0, NULL, 'h'},
		{"ports", 1, NULL, 'p'},
		{"tty", 1, NULL, 't'},
		{"watch", 1, NULL, 'w'},
		/* {"wait", 0, NULL, 0}, */ /* Disabled.  Probably not useful. */
		{NULL, 0, NULL, 0}
	};

	while ((opt = getopt_long(argc, argv, "dhp:t:w?", long_options, &option_index)) != -1) {
		switch (opt) {
			case 0:
				for (i = 0; i < (sizeof(R_commands) / sizeof(struct R_command)); i++) {
					if (!R_commands[i].option || strcmp(R_commands[i].option, long_options[option_index].name)) {
						continue;
					}

					cmd_mask |= (1 << i);
					if (optarg) {
						optargs[i] = strdup(optarg);
					}
				}
				break;
			case 'd':
				debugmode = 1;
				break;
			case 'p':
			{
				char *port = NULL;

				/* This might be useful some day.  Probably not. */
				if (!strcmp(optarg, "all")) {
					ports = 0xFFFF;
					break;
				} else if (!strcmp(optarg, "odd")) {
					ports = 0x5555;
					break;
				} else if (!strcmp(optarg, "even")) {
					ports = 0xAAAA;
					break;
				}

				while (port = strsep(&optarg, ",")) {
					if (atoi(port) > 0) {
						ports |= 1 << (atoi(port) - 1);
					}
				}
				break;
			}
			case 't':
				if (usbtty) {
					R_err("USB TTY already set.  Using '%s' instead.\n", optarg);
				}
				usbtty = strdup(optarg);
				break;
			case 'w':
				timeout = -1;
				break;
			case 'h':
			case '?':
			default:
				R_usage();
				exit(1);
				break;
		}
	}

	ttyfd = open(usbtty, O_RDWR | O_NOCTTY | O_NONBLOCK);
	if (ttyfd == -1) {
		R_err("Could not open tty %s: %s\n", usbtty, strerror(errno));
		exit(1);
	}

	bzero(&termios_p, sizeof(termios_p)); /* clear struct for new port settings */
	termios_p.c_cflag = B115200 | CS8 | CLOCAL | CREAD;
	termios_p.c_iflag = IGNBRK | IGNPAR | IXON;
	termios_p.c_oflag = 0;
	termios_p.c_lflag = 0;
	termios_p.c_cc[VMIN] = 1;
	termios_p.c_cc[VTIME] = 0;
	tcflush(ttyfd, TCIOFLUSH); /* Clear input/output buffers */
	tcsetattr(ttyfd, TCSANOW, &termios_p);


	/* Keep this at the beginning. */
	if (cmd_mask & (1 << CMD_FIRMWARE_STAGE1) && cmd_mask & (1 << CMD_FIRMWARE_STAGE2)) {
		/* Do not execute any other requests. */
		cmd_mask = (1 << CMD_FIRMWARE_STAGE1) | (1 << CMD_FIRMWARE_STAGE2);

		firmware = optargs[CMD_FIRMWARE_STAGE2];

		request_firmware_stage1();
		request_firmware_stage2();
	}

	if (cmd_mask & (1 << CMD_GET_FIRMWARE_VERSION)) {
		request_get_firmware_version();
	}
	if (cmd_mask & (1 << CMD_GET_SERIAL_NUMBER)) {
		request_get_serial_number();
	}
	if (cmd_mask & (1 << CMD_SET_SERIAL_NUMBER)) {
		request_set_serial_number(optargs[CMD_SET_SERIAL_NUMBER]);
	}
	if (cmd_mask & (1 << CMD_GET_PRODUCT_NUMBER)) {
		request_get_product_number();
	}
	if (cmd_mask & (1 << CMD_SET_PRODUCT_NUMBER)) {
		request_set_product_number(optargs[CMD_SET_PRODUCT_NUMBER]);
	}
	if (cmd_mask & (1 << CMD_GET_PORT_COUNT)) {
		request_get_port_count();
	}
	if (cmd_mask & (1 << CMD_GET_CURRENT_CTLPORT)) {
		request_get_current_ctlport();
	}
	if (cmd_mask & (1 << CMD_GET_MASTER_CTLPORT)) {
		request_get_master_ctlport();
	}
	if (cmd_mask & (1 << CMD_GET_ID_SWITCH)) {
		request_get_id_switch();
	}
	if (cmd_mask & (1 << CMD_GET_WATCHDOG)) {
		request_get_watchdog();
	}
	if (cmd_mask & (1 << CMD_SET_WATCHDOG)) {
		request_set_watchdog(atoi(optargs[CMD_SET_WATCHDOG]));
	}
	if (cmd_mask & (1 << CMD_PING_WATCHDOG)) {
		request_ping_watchdog();
	}
	if (cmd_mask & (1 << CMD_CONTROL_REQUEST)) {
		request_control_request();
	}
	if (cmd_mask & (1 << CMD_CONTROL_RELEASE)) {
		control_request_release();
	}
	if (cmd_mask & (1 << CMD_SET_CONTROL_MODE)) {
		request_set_control_mode(atoi(optargs[CMD_SET_CONTROL_MODE]));
	}
	if (cmd_mask & (1 << CMD_GET_CONTROL_MODE)) {
		request_get_control_mode();
	}
	if (cmd_mask & (1 << CMD_SET_PORT_STATUS)) {
		request_set_port_status(atoi(optargs[CMD_SET_PORT_STATUS]));
	}
	if (cmd_mask & (1 << CMD_GET_PORT_STATUS)) {
		request_get_port_status();
	}

	/* Kludge to watch all responses. */
	if (timeout == -1) {
		cmd_mask = 0xFFFFFFFF;
	}

	fds.fd = ttyfd;
	fds.events = POLLIN | POLLPRI;
	while (cmd_mask > 0 && poll(&fds, 1, timeout) > 0) {
		R_debug("poll() returned\n");
		if (!fds.revents) {
			break;
		}
		if (handle_response()) {
			break;
		}
	}
	if (cmd_mask > 0 && cmd_mask != 0xFFFFFFFF) {
		R_debug("cmd_mask was non-zero after poll() timeout.\n");
		if (cmd_mask & (1 << CMD_FIRMWARE_STAGE1) && cmd_mask & (1 << CMD_FIRMWARE_STAGE2)) {
			update_firmware();
		} else {
			exit(1);
		}
	}
	close(ttyfd);
	exit(0);
}
