/*
 * rctl -- Digium R-Series control application
 *
 * Copyright (C) 2010-2011, Digium, Inc.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2.  See the COPYING file
 * at the top of the source tree.
 */

/*
 * Serial port driver for LPC17xx processor ISP protocol
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <errno.h>
#include <string.h>

#include "uu.h"

#define CMD_RETRIES			(20)
#define SYNCH_ATTEMPTS		(10)
#define SYNCH_STRING		"Synchronized"
#define OK_STRING			"OK"
#define RESEND_STRING		"RESEND"

#define LINES_PER_CHECKSUM	(20)
#define DEFAULT_US_WAIT		(20000)
#define DATA_RESPONSE_TRIES	(8)

/* ISP command responses explicitly checked for */
#define SECTOR_NOT_BLANK	(8)
#define COMPARE_ERROR		(10)

/* FLASH related constants */
#define SIZE4K_SECTOR 	0x1000
#define SIZE32K_SECTOR 	0x8000
#define NUM4K_SECTORS 	16
#define NUM32K_SECTORS 	14

/* User application FLASH segmentation */
#define TARGET_IMAGE_START		(0xd000)
#define TARGET_IMAGE_END		(0x7FFFF)	/* LPC1758 FLASH size */
//#define TARGET_IMAGE_END		(0xFFFF)  /* LPC1752 FLASH size */
#define TARGET_HEADER_SIZE		(0x100)
#define TARGET_PROGRAM_START	(TARGET_IMAGE_START + TARGET_HEADER_SIZE)
#define TARGET_PROGHEADER_START	(TARGET_IMAGE_START)

/* Parameters for locations on the host processor that can be used to stage
 * data before it is written to FLASH.  Values chosen for no other reason than
 * that they work. 
 */ 
//#define FLASH_CHUNK_SIZE	(512)
#define FLASH_CHUNK_SIZE	(896) /* Must be divisible by 4, and less than (45 * 20) */
#define FLASH_CHUNK_RAMADDR	(0x10002000)

/* The AES Initialization vector size */
#define AES_IV_SIZE			(32)

/* Static variables describing a single session */
static int echo_state;		/* indicator if ISP is echoing input */

/* structure to provide text responses for error codes */
struct response_error {
	int code;
	char * codemsg;
};

struct response_error isp_response_codes[] = {
				{0, "CMD_SUCCESS"},
				{1, "INVALID_COMMAND"},
				{2, "SRC_ADDR_ERROR"},
				{3, "DST_ADDR_ERROR"},
				{4, "SRC_ADDR_NOT_MAPPED"},
				{5, "DST_ADDR_NOT_MAPPED"},
				{6, "COUNT_ERROR"},
				{7, "INVALID_SECTOR"},
				{8, "SECTOR_NOT_BLANK"},
				{9, "SECTOR_NOT_PREPARED_FOR_WRITE_OPERATION"},
				{10, "COMPARE_ERROR"},
				{11, "BUSY"},
				{12, "PARAM_ERROR"},
				{13, "ADDR_ERROR"},
				{14, "ADDR_NOT_MAPPED"},
				{15, "CMD_LOCKED"},
				{16, "INVALID_CODE"},
				{17, "INVALID_BAUD_RATE"},
				{18, "INVALID_STOP_BIT"},
				{19, "CODE_READ_PROTECTION_ENABLED"},
				{20, "DECRYPTION_ERROR"},
				{-1, ""}}; /* Terminates the array */

char inbuffer[4096];		/* Buffer of raw input taken in from the serial port */
int read_mark;				/* Marker where next data goes */
extern int ttyfd;

/* -----------------------------------------------------------------------------
 * read_serport
 *
 * @param  buf - buffer to place returned command .
 *	   space - number of bytes available in the buffer
 * @return Number of bytes in the response or <0 is an error
 *
 * @brief  Reads data in from the serial port and stores into a stores in a local
 *	buffer.  It then looks for a full command in the local buffer.  If som the
 *	command is returned to the calling function
 *
 * ----------------------------------------------------------------------------
 */
static int read_serport(char * buf, int space)
{
	int numread;
	int ret = 0;

	/* sanity check */
	if (NULL == buf || 0 >= space) {
		return -1;
	}

	buf[0] = '\0'; /* Always init read buffer */

	numread = read(ttyfd, &inbuffer[read_mark], sizeof(inbuffer) - read_mark -1);
	if (0 > numread) {
		if (EAGAIN == errno) {
			numread = 0;
		} else {
			R_err("Serial Read error %d (%d, %s)\n", numread, errno, strerror(errno));
			return -1;
		}
	} else if (0 < numread) {
		read_mark += numread;
		inbuffer[read_mark] = '\0';
	}
	/* Determine if full command is in buffer */
	if (strchr(inbuffer, '\n')) {
		int i,j;
		for (i = 0, j = 0; j < space-1 && inbuffer[i] != '\n'; i++) {
			/* Throw away carriage returns, we only use newline */
			if (inbuffer[i] != '\r') {
				buf[j++] = inbuffer[i];
			}
		}
		if (j < space-2) {
			buf[j++] = '\n';
		}
		buf[j] = '\0';
		ret = j;
		/* Rearrange the incoming data */
		for (j = 0, i++; i < read_mark; i++, j++) {
			inbuffer[j] = inbuffer[i];
		}
		inbuffer[j] = '\0';
		read_mark = j;
	}
	return ret;
}

/* -----------------------------------------------------------------------------
 * write_serport
 *
 * @param	msg - ascii string to be written to the serial port.
 * @return number of characters written or < 0 == error 
 *
 * @brief 	Write data out the serial port
 *
 * ----------------------------------------------------------------------------
 */
static int write_serport(char * msg)
{
	int len = strlen(msg);
	int res;

	if (ttyfd) {
		res = write(ttyfd, msg, len);
		if (0 > res) {
			R_err("Serial Write error %d\n", res);
			return -1;

		}
		return len;
	}
	return -1;
}

/* -----------------------------------------------------------------------------
 * flush_serport
 *
 * @param
 * @return 
 *
 * @brief  Read all data out of the serial port and then reset the input buffer 
 *
 * ----------------------------------------------------------------------------
 */
static void flush_serport(void)
{
	if (ttyfd) {
		while (0 < read(ttyfd,&inbuffer[0],sizeof(inbuffer)))
		{}
	}
	bzero(inbuffer, sizeof(inbuffer));
	read_mark = 0;
}

/* -----------------------------------------------------------------------------
 * isp_response_string
 *
 * @param code to map 
 * @return pointer to character string matching code 
 *
 * @brief - Map a code to an string describing the code. 
 *
 * ----------------------------------------------------------------------------
 */
char * isp_response_string(int code)
{
	char * resp = "Unknown";
	int idx;

	for (idx = 0; 0 <= isp_response_codes[idx].code; idx++) {
		if (code == isp_response_codes[idx].code) {
			resp = isp_response_codes[idx].codemsg;
			break;
		}
	}
	return resp;
}

/* -----------------------------------------------------------------------------
 * isp_get_sectornumber
 *
 * @param Memory address on the NXP processor
 * @return The sector number in which the address resides or -1 if no flash page exists
 *
 * @brief  NOTE: function assumes largest FLASH configuration available,
 *	The sector returned may not reside on devices with restricted memory space.
 *
 * ----------------------------------------------------------------------------
 */
int isp_get_sectornumber( uint32_t address)
{
	int sector = -1;

	if (address < SIZE4K_SECTOR * NUM4K_SECTORS) {
		sector = address / SIZE4K_SECTOR;
	}
	else if (address < SIZE4K_SECTOR * NUM4K_SECTORS + SIZE32K_SECTOR * NUM32K_SECTORS) {
		address -= SIZE4K_SECTOR * NUM4K_SECTORS;
		sector = NUM4K_SECTORS + address / SIZE32K_SECTOR;
	}

	return sector;
}

/* -----------------------------------------------------------------------------
 * isp_get_startaddress_from_sector
 *
 * @param	sector number
 * @return The first address that resides in the sector or -1 if there is no
 *		such address
 *
 * @brief
 *
 * ----------------------------------------------------------------------------
 */
static uint32_t isp_get_startaddress_from_sector( int sector)
{
	uint32_t address = (uint32_t)-1;;
	if (0 <= sector && NUM4K_SECTORS + NUM32K_SECTORS > sector) {
		if (sector < NUM4K_SECTORS) {
			address = sector * SIZE4K_SECTOR;
		} else {
			address = (sector - NUM4K_SECTORS) * SIZE32K_SECTOR + NUM4K_SECTORS * SIZE4K_SECTOR;
		}
	}
	return address;
}

/* -----------------------------------------------------------------------------
 * isp_get_endaddress_from_sector
 *
 * @param sector number
 * @return The last address that resides in the sector or -1 if there is no
 *		such address
 *
 * @brief
 *
 * ----------------------------------------------------------------------------
 */
static uint32_t isp_get_endaddress_from_sector( int sector)
{
	uint32_t address = (uint32_t)-1;;
	if (0 <= sector && NUM4K_SECTORS + NUM32K_SECTORS > sector) {
		if (sector < NUM4K_SECTORS) {
			address = ((sector + 1) * SIZE4K_SECTOR) - 1;
		} else {
			address = (sector - NUM4K_SECTORS + 1) * SIZE32K_SECTOR -1 + NUM4K_SECTORS * SIZE4K_SECTOR ;
		}
	}
	return address;
}

/* -----------------------------------------------------------------------------
 * isp_doSimpleCmd
 *
 * @param cmd - command to be sent to ISP host processor
 * @return > 0 - response code returned from host, < 0 - error
 *
 * @brief  Sends a command to the host and then waits for a response.in the
 *	form of a response code.
 *	Any data sent after the response code is the responsibility of the
 *	calling function.
 *
 * ----------------------------------------------------------------------------
 */
static int isp_doSimpleCmd(char *cmd)
{
	int	attempts;
	int	resplen;
	int	step;
	int retries = 0;
	int	respcode = -1;
	char	cmdresp[64];

	write_serport(cmd);
	step = echo_state ? 0 : 1;
	for (attempts = 0; 0 > respcode && attempts < CMD_RETRIES; attempts++) {
		usleep(DEFAULT_US_WAIT);
		resplen = read_serport(cmdresp, sizeof(cmdresp));
		if (0 < resplen) {
			switch (step) {
			case 0:
				if (!strcmp(cmd, cmdresp)) {
					step++;
					attempts = -1;
				}
				break;
			case 1:
				respcode = strtoul(cmdresp, NULL, 10);
				if (respcode == 1 && retries < 1) {
					retries++;
					write_serport(cmd);
					attempts = -1;
					step = echo_state ? 0 : 1;
				} else {
					step++;
				}
				break;
			}
		}
	}
	return respcode;
}

/* -----------------------------------------------------------------------------
 * isp_synch
 *
 * @param
 * @return 0 = success, else error
 *
 * @brief - Performs the tasks required to synchronize to tghe host processor.
 *	The host processor comes up in an autobaud configuration in which it
 *	expects to see a "?" character.   Upon receipt, the host determines the
 *	baud rate of the interface.  It then passes back and forth a "Synchronized"
 * 	message to confirm the baud rate selection. 
 *
 * See:  32.3.1.1 Criterion for Valid User Code in UM10360 ("LPC17xx User manual")
 * ----------------------------------------------------------------------------
 */
int isp_synch(void)
{
	char response[64];
	int attempts;
	int resplen;
	int step = 0;

	for (attempts = 0; attempts < SYNCH_ATTEMPTS; attempts++) {
		write_serport("?");
		usleep(500000);
		resplen = read_serport(response, sizeof(response));
		if (strlen(SYNCH_STRING) <= resplen) {
			if (NULL != strstr(response, SYNCH_STRING)) {
				R_log("Synching...\n");
				write_serport("Synchronized\n");
				step++;
				break;
			}
		}
	}
	if (0 < step) {
		for (attempts = 0; attempts < SYNCH_ATTEMPTS; attempts++) {
			usleep(500000);
			resplen = read_serport(response, sizeof(response));
			if (strlen(OK_STRING) <= resplen) {
				if (NULL != strstr(response, OK_STRING)) {
					R_log("Confirmed\n");
					step++;
					break;
				}
			}
		}
	}
	/* I've synched to the host, send confirming new line and clear out any buffering */
	if (1 < step) {
		write_serport("\n\n\n");
		echo_state = 1;
		// Assume echo for now write_serport("A 0\n"); /* No echo */
		usleep(500000);
		flush_serport();
		step++;
	}

	return step >= 2 ? 0 : -1;
}

/* -----------------------------------------------------------------------------
 * isp_setEcho
 *
 * @param 0 - Turn echo off, otherwise on
 * @return echo state after attempting to change it. 
 *
 * @brief
 *
 * See:  32.7.3 Echo <setting> in UM10360 ("LPC17xx User manual")
 * ----------------------------------------------------------------------------
 */
int isp_setEcho(int state)
{
	char msg[8];

	snprintf(msg, sizeof(msg), "A %d\n", state);
	if (!isp_doSimpleCmd(msg)) {
		echo_state = state;
	} else {
		/* It failed but this might be due to the present echo state already being the desired state */
		echo_state = !echo_state;
		if (!isp_doSimpleCmd(msg)) {
			echo_state = state;
		} else {
			echo_state = !echo_state;	/* Nope, give up and put it back to original */
		}
	}
	return echo_state;
}

/* -----------------------------------------------------------------------------
 * isp_doW - Memory Write command 
 *
 * @param address - RAM location to write data to
 *		len - number of bytes 
 * @return 0 = success, else error
 *
 * @brief Write data to memory.  All data should be on 4 byte word boundaries
 *
 * See: 32.7.4 Write to RAM <start address> <number of bytes> in UM10360 ("LPC17xx User manual")
 * ----------------------------------------------------------------------------
 */
int isp_doW(uint32_t address, int len, uint8_t *buffer)
{
	char cmd[32];
	int cmdresp;
	char uu_buffer[128];
	uint8_t uu_raw[48];
	char datalinecount;
	int i, j;
	uint32_t checksum_calculated;
	int marker = 0;
	int lastmarker = 0;
	int checksummarker = 0;

	if (address & 3) {
		R_err("W Command: Address (0x%x) must be 4 byte aligned\n", address);
		return -1;
	}
	if (0 >= len || len & 3 ) {
		R_err("W Command: Length (0x%x) must be 4 byte aligned\n", len);
		return -1;
	}

	snprintf(cmd, sizeof(cmd), "W %d %d\n", address, len);
	cmdresp = isp_doSimpleCmd(cmd);
	datalinecount = 0;
	if (0 != cmdresp) {
		R_err("The W command response is not right: %d - %s\n",
							cmdresp,
							isp_response_string(cmdresp));
	}

	while (marker < len) {
		j = len - marker > 45 ? 45 : len - marker;
		memcpy(uu_raw, &buffer[marker], j);
		i = uuencode(uu_raw, j, uu_buffer);
		write_serport(uu_buffer);
		if (echo_state) {
			usleep(DEFAULT_US_WAIT);
			read_serport(uu_buffer, sizeof(uu_buffer));
		}
		marker += j;
		datalinecount++;

		/* See if it is time to generate checksum for verification */
		if (LINES_PER_CHECKSUM <= datalinecount || marker == len) {
			for (checksummarker = lastmarker, checksum_calculated = 0;
					checksummarker < marker;
					checksummarker++) {
				checksum_calculated += buffer[checksummarker];
			}
			snprintf(cmd, sizeof(cmd), "%d\n", checksum_calculated);
			write_serport(cmd);
			if (echo_state) {
				usleep(DEFAULT_US_WAIT);
				read_serport(uu_buffer, sizeof(uu_buffer));
			}
			i = 0;
			while (1) {
				usleep(DEFAULT_US_WAIT);
				if (0 < read_serport(uu_buffer, sizeof(uu_buffer))) {
					break;
				}
				if (++i > DATA_RESPONSE_TRIES) {
					R_err("W command: No Confirmation response. Aborting\n");
					return -1;
				}
			}
			if (!strcmp("OK\n", uu_buffer)) {
				lastmarker = marker;
			} else if (!strcmp("RESEND\n", uu_buffer)) {
				marker = lastmarker;
				R_err("Resending data on checksum error\n");
			} else {
				R_err("W command: Unknown response (%s). Aborting\n", uu_buffer);
				return -1;
			}
			datalinecount = 0;
		}
	}
	return 0;
}

/* -----------------------------------------------------------------------------
 * isp_doG - Execute Program
 *
 * @param address - address to jump toYou
 * @return 0 if valid, else error response code
 *
 * @brief
 *
 * See: 32.7.8 Go <address> <mode> in UM10360 ("LPC17xx User manual")
 * ----------------------------------------------------------------------------
 */
int isp_doG(uint32_t address)
{
	char cmd[32];
	int cmdresp;

	if (isp_doU()) {
		R_err("G Cmd: Unable to unlock to execute\n");
		return -1;
	}

	snprintf(cmd, sizeof(cmd), "G %d T\n", address);
	cmdresp = isp_doSimpleCmd(cmd);
	if (-1 == cmdresp) {
		cmdresp = 0;
	}
	return cmdresp;
}

/* -----------------------------------------------------------------------------
 * isp_doP - Prepare for FLASH erase/write 
 *
 * @param start address, end address - 
 * @return 0 = success, else error
 *
 * @brief Prepares all the sectors touched by the address range specified 
 *
 * See:  32.7.6 Prepare sector(s) for write operation <start sector number> <end
 * 	sector number> in UM10360 ("LPC17xx User manual")
 * ----------------------------------------------------------------------------
 */
int isp_doP(uint32_t start, uint32_t end)
{
	char cmd[64];
	int cmdresp;
	int start_sector = isp_get_sectornumber(start);
	int end_sector = isp_get_sectornumber(end);

	if (0 > start_sector){
		R_err("P Cmd: Bad FLASH start address - 0x%x\n", start);
		return -1;
	}
	if (0 > end_sector){
		R_err("P Cmd: Bad FLASH end address - 0x%x\n", end);
		return -1;
	}
	if (start > end){
		R_err("P Cmd: Start (0x%x) must be less than end (0x%x)\n", start, end);
		return -1;
	}

	snprintf(cmd, sizeof(cmd), "P %d %d\n", start_sector, end_sector);
	cmdresp = isp_doSimpleCmd(cmd);
	if (0 != cmdresp) {
		R_err("P cmd: Prepare for sectors %d to %d failed.  %d - %s\n",
				start_sector, end_sector,
				cmdresp, isp_response_string(cmdresp));
	}
	return cmdresp;
}

/* -----------------------------------------------------------------------------
 * isp_doC_orX - Copy RAM to FLASH 
 *
 * @param ram base address
 *	flash address
 *	len - number of bytes to copy
 *
 * @brief Performs the write to FLASH from RAM memory.  Requires Unlock and Prepare
 *	commands to precede it. 
 *
 * See: i32.7.7 Copy RAM to Flash <flash address> <RAM address> <no of bytes>
 * 	in UM10360 ("LPC17xx User manual")
 * ----------------------------------------------------------------------------
 */

int isp_doC_or_X(uint32_t ram, uint32_t flash, int len, int dodecrypt)
{
	char cmd[64];
	int cmdresp;
	int startsector = isp_get_sectornumber(flash);
	int endsector = isp_get_sectornumber(flash+len-1);
	char * cmdtype = "C";

	if (dodecrypt) {
		cmdtype = "X";
	}

	if (0 > startsector || 0 > endsector){
		R_err("%s Cmd: Bad FLASH address - 0x%x\n", cmdtype, flash);
		return -1;
	}

	if (isp_doU()) {
		R_err("%s Cmd: Unable to unlock FLASH for writing\n", cmdtype);
		return -1;
	}
	if (isp_doP(isp_get_startaddress_from_sector(startsector), isp_get_endaddress_from_sector(endsector))) {
		R_err("%s Cmd: Unable to prepare FLASH for writing\n", cmdtype);
		return -1;
	}

	snprintf(cmd, sizeof(cmd), "%s %d %d %d\n", cmdtype, flash, ram, len);
	cmdresp = isp_doSimpleCmd(cmd);
	if (0 != cmdresp) {
		R_err("%s cmd: Flash write for sectors %d-%d (address 0x%x to 0x%x) failed.  %d - %s\n",
				cmdtype, startsector, endsector, flash, flash + len -1,
				cmdresp, isp_response_string(cmdresp));
	}
	return 0;
}

/* -----------------------------------------------------------------------------
 * isp_doE - Erase FLASH sectors
 *
 * @param Start - Start address
 *	End - end address
 * @return 0 = success, else error
 *
 * @brief  Erases sectors of FLASH memory.  Requires Unlock and Prepare
 *	commands to precede it. 
 *
 * See: 32.7.9 Erase sector(s) <start sector number> <end sector number>
 * 	in UM10360 ("LPC17xx User manual")
 * ----------------------------------------------------------------------------
 */
int isp_doE(uint32_t start, uint32_t end)
{
	char cmd[64];
	int cmdresp;
	int start_sector = isp_get_sectornumber(start);
	int end_sector = isp_get_sectornumber(end);

	if (0 > start_sector){
		R_err("E Cmd: Bad FLASH start address - 0x%x\n", start);
		return -1;
	}
	if (0 > end_sector){
		R_err("E Cmd: Bad FLASH end address - 0x%x\n", end);
		return -1;
	}
	if (start > end){
		R_err("E Cmd: Start (0x%x) must be less than end (0x%x)\n", start, end);
		return -1;
	}
	if (isp_doU()) {
		R_err("E Cmd: Unable to unlock FLASH for erasure\n");
		return -1;
	}
	if (isp_doP(isp_get_startaddress_from_sector(start_sector), isp_get_endaddress_from_sector(end_sector))) {
		return -1;
	}

	snprintf(cmd, sizeof(cmd), "E %d %d\n", start_sector, end_sector);
	cmdresp = isp_doSimpleCmd(cmd);
	if (0 != cmdresp) {
		R_err("E cmd: Erase from sectors %d to %d failed.  %d - %s\n",
				start_sector, end_sector,
				cmdresp, isp_response_string(cmdresp));
	}
	return cmdresp;
}

/* -----------------------------------------------------------------------------
 * isp_doU - unlock command
 *
 * @param
 * @return 0 = success, else error
 *
 * @brief
 *
 * See: 32.7.1 Unlock <Unlock code> in UM10360 ("LPC17xx User manual")
 * ----------------------------------------------------------------------------
 */
int isp_doU(void)
{
	int cmdresp = isp_doSimpleCmd("U 23130\n");
	if (0 != cmdresp) {
		R_err("U cmd: Unlock failed.  %d - %s\n",
				cmdresp, isp_response_string(cmdresp));
				return -1;
	}
	return 0;
}

/* -----------------------------------------------------------------------------
 * isp_doV - Load the AES initialization vector and encrypted header
 *
 * @param iv - pointer to the initialization data
 * @return 0 = success, else error
 *
 * @brief
 *
 * This is a non-standard ISP command developed specifically for the bootloader
 * ----------------------------------------------------------------------------
 */
int isp_doV(uint8_t * iv)
{
	char cmd[64];
	int cmdresp;

	if (isp_doW(FLASH_CHUNK_RAMADDR, AES_IV_SIZE, iv)) {
		R_err("W Cmd: Unable write iv vector\n");
		return -1;
	}
	snprintf(cmd, sizeof(cmd), "V %d\n", FLASH_CHUNK_RAMADDR);
	cmdresp = isp_doSimpleCmd(cmd);
	if (0 != cmdresp) {
		R_err("V cmd: IV Vector set failed.  %d - %s\n",
				cmdresp, isp_response_string(cmdresp));
	}
	return cmdresp;
}

/* -----------------------------------------------------------------------------
 * isp_doH - Set the Header for an image in memory
 *
 * @param length - The length of the image that is to be verified by the target
 *		version - version number that gets stored with the image header
 *		id - character string that gets stored in the header
 * @return 0 = success, else error
 *
 * @brief  Prompts the target to validate the image that has been stored in FLASH.
 *			If the image validates a header gets stored in FLASH and the image is
 *			executed upon the next reboot.
 *
 * This is a non-standard ISP command developed specifically for the bootloader
 * ----------------------------------------------------------------------------
 */
int isp_doH(unsigned length, unsigned version, char * id)
{
	char cmd[64];
	int cmdresp;

	snprintf(cmd, sizeof(cmd), "H %d %d, %-16.16s\n", length, version, "fred" /*id*/);
	cmdresp = isp_doSimpleCmd(cmd);
	if (0 != cmdresp) {
		R_err("H cmd: Setting header failed %d - %s\n",
				cmdresp, isp_response_string(cmdresp));
	}
	return cmdresp;
}

/* -----------------------------------------------------------------------------
 * write_enc_block - Writes an encrypted image to FLASH
 *
 * @param  address - starting address in FLASH to write
 * 	len - number of bytes to write
 *	buffer - data to be written
 * @return 0 = success, else error
 *
 * @brief  - Writes an image that is encrypted to user application FLASH
 * 		This is meant to interface to the bootloader ISP interface.
 * 		Assumes that User App FLASH is erased
 *
 * ----------------------------------------------------------------------------
 */
int write_enc_block(uint32_t address, int len, uint8_t * buffer)
{
	int blocksize;

	/* While there is data to be programmed */
	while (0 < len) {
		/* Determine how big a block can be programmed now */
		if (len < FLASH_CHUNK_SIZE) {
			blocksize = len;
		} else {
			blocksize = FLASH_CHUNK_SIZE;
		}
		/* Write the block to RAM */
		if (isp_doW(FLASH_CHUNK_RAMADDR, blocksize, buffer)) {
			R_err("Cannot write to RAM memory\n");
			return -1;
		}
		/* Decrypt is and then copy to FLASH */
		if (isp_doC_or_X(FLASH_CHUNK_RAMADDR, address, blocksize, 1)) {
			R_err("Cannot write RAM to FLASH memory\n");
			return -1;
		}
		buffer += blocksize;
		address += blocksize;
		len -= blocksize;
	}
	return 0;
}


/* -----------------------------------------------------------------------------
 * writeEncryptedImage - Writes an encrypted image to FLASH
 *
 * @param  buffer - data to be written
 * 			len - Number of bytes to write
 * 			version - Version of the image being programmed
 * 			id - Identification of the image being programmed
 * @return 0 = success, else error
 *
 * @brief  -
 * 	Writes an encrypted image to FLASH.  For use in interfacing the bootloader ISP
 * 	Takes into account the overlap of the actual image and the data
 *	spillover in the FLASH sectors.
 *
 * ----------------------------------------------------------------------------
 */
int writeEncryptedImage(uint8_t * buffer, int len, unsigned version, char *id)
{

	if (isp_doV(buffer)) {
		R_err("Could not send Decryption Information\n");
		return -1;
	}
	buffer 	+= AES_IV_SIZE;
	len 	-= AES_IV_SIZE;

	if (isp_doE(TARGET_IMAGE_START,TARGET_IMAGE_END)) {
		R_err("Could not Erase the Application FLASH area\n");
		return -1;
	}
	if (write_enc_block(TARGET_PROGRAM_START, len, buffer)) {
		R_err("Could not program image of size %d to application FLASH\n", len);
		return -1;
	}
	if (isp_doH(len, version, id)) {
		R_err("Could not Set Header for program image\n");
		return -1;
	}
	return 0;
}
