/*
 * rctl -- Digium R-Series control application
 *
 * Copyright (C) 2010-2011, Digium, Inc.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2.  See the COPYING file
 * at the top of the source tree.
 */

/*
 * UU encoding and decoding functions.
 */

#include "uu.h"

/* -----------------------------------------------------------------------------
 * UU decode
 *
 * @param  String pointer to be decoded, buffer in which decoded data is placed
 * @return Number of characters interpreted
 *
 * @brief  Takes a UUEncoded string and decodes it into the provided buffer
 * 			The buffer needs to be of the size indicated by the first byte in
 * 			the string rounded out to where the number is divisible by 3.
 * 			The '`' character replaces the ' ' character in strings
 * ----------------------------------------------------------------------------
 */
int uudecode(char *asciimsg, unsigned char *buffer)
{
	int i;
	char rawchars[4];
	int done = 0;
	int subcount;
	int calcnum = 0;

	unsigned char * decode = buffer;

	/* Check for 0 length string */
	if (*asciimsg == '`') {
		*asciimsg = ' ';
	}
	/* Determine number of bytes encoded in string from first byte */
	calcnum = *asciimsg++ - ' ';

	while(!done) {
		/* look through uuencoded bytes, 4 bytes at a time
		 *  subtract the 0x020 so that LSB 6 bits contain the data to be decoded
		 */
		for (i = 0, subcount = 0; i < 4; i++) {
			if (*asciimsg == '\0' || *asciimsg == '\n') {
				rawchars[i] = 0;
				done = 1;
			} else {
				if ('`' == *asciimsg) {
					*asciimsg = ' ';
				}
				rawchars[i] = *asciimsg++ - 0x20;
				subcount++;
			}
		}
		/* Decode the 4 bytes into 3 bytes of binary data */
		if (0 < subcount) {
			*decode++ = ((rawchars[0] & 0x3f) << 2) | ((rawchars[1] & 0x30) >> 4);
		}
		if (1 < subcount) {
			*decode++ = ((rawchars[1] & 0x0f) << 4) | ((rawchars[2] & 0x3c) >> 2);
		}
		if (2 < subcount) {
			*decode++ = ((rawchars[2] & 0x03) << 6) | ((rawchars[3] & 0x3f) << 0);
		}
		/* Check to see if the string has been decoded?  If so, get out */
		if (*asciimsg == '\0' || *asciimsg == '\n') {
			done = 1;
		}
	}
	return calcnum;
}


/* -----------------------------------------------------------------------------
 * UU encode
 *
 * @param  buffer holding data to be encoded,
 * 			the # of bytes to encode
 * 			The buffer to place the string.
 * @return Number of characters interpreted
 *
 * @brief  UU Encodes a binary buffer to an ASCII string.
 * 			The string buffer needs to be of the size
 * 						START BYTE + (num * 4 / 3) + Linefeed
 * 			The '`' character replaces the ' ' character in strings
 * ----------------------------------------------------------------------------
 */
int uuencode(unsigned char *buffer, int num, char * inbuffer)
{
	int count = 0;
	int deccount = 0;
	char * asciibuffer = inbuffer;

	if (num == 0) {
		inbuffer[0] = '`';
		inbuffer[1] = '\n';
		return 0;
	}

	*asciibuffer++ = ' ' + num;

	if (num % 3 == 1) {
		buffer[num++] = 0;
	}
	if (num % 3 == 2) {
		buffer[num++] = 0;
	}

	for (count = 0; count < num; count +=3) {
		*asciibuffer =  ((buffer[0] & 0xfc) >> 2) + 0x20;
		if (*asciibuffer == ' ') *asciibuffer = '`';
		asciibuffer++;
		*asciibuffer = (((buffer[0] & 0x03) << 4) | (buffer[1] & 0xf0) >> 4) + 0x20;
		if (*asciibuffer == ' ') *asciibuffer = '`';
		asciibuffer++;
		*asciibuffer = (((buffer[1] & 0x0f) << 2) | (buffer[2] & 0xc0) >> 6) + 0x20;
		if (*asciibuffer == ' ') *asciibuffer = '`';
		asciibuffer++;
		*asciibuffer =   (buffer[2] & 0x3f) + 0x20;
		if (*asciibuffer == ' ') *asciibuffer = '`';
		asciibuffer++;
		deccount += 4;
		buffer += 3;
	}
	*asciibuffer++ = '\n';
	*asciibuffer = '\0';
	return deccount;
}
